@@ -13,7 +13,7 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 # 
-# RCS: @(#) $Id: ftp.tcl,v 1.12 2001/07/10 20:39:46 andreas_kupries Exp $
+# RCS: @(#) $Id: ftp.tcl,v 1.23 2002/01/18 20:51:16 andreas_kupries Exp $
 #
 #   core ftp support: 	ftp::Open <server> <user> <passwd> <?options?>
 #			ftp::Close <s>
@@ -26,26 +26,41 @@
 #			ftp::ModTime <s> <from> <to>
 #			ftp::Delete <s> <file>
 #			ftp::Rename <s> <from> <to>
-#			ftp::Put <s> <(local | -data "data"> <?remote?>
-#			ftp::Append <s> <(local | -data "data"> <?remote?>
-#			ftp::Get <s> <remote> <?(local | -variable varname)?>
+#			ftp::Put <s> <(local | -data "data" -channel chan)> <?remote?>
+#			ftp::Append <s> <(local | -data "data" | -channel chan)> <?remote?>
+#			ftp::Get <s> <remote> <?(local | -variable varname | -channel chan)?>
 #			ftp::Reget <s> <remote> <?local?>
 #			ftp::Newer <s> <remote> <?local?>
 #			ftp::MkDir <s> <directory>
 #			ftp::RmDir <s> <directory>
 #			ftp::Quote <s> <arg1> <arg2> ...
 #
+# Internal documentation. Contents of a session state array.
+#
+# ---------------------------------------------
+# key             value
+# ---------------------------------------------
+# State           Current state of the session and the currently executing command.
+# RemoteFileName  Name of the remote file, for put/get
+# LocalFileName   Name of local file, for put/get
+# inline          1 - Put/Get is inline (from data, to variable)
+# filebuffer  
+# PutData         Data to move when inline
+# SourceCI        Channel to read from, "Put"
+# ---------------------------------------------
+#
 
-package provide ftp [lindex {Revision: 2.2 } 1]
+package require Tcl 8.2
+package require log     ; # tcllib/log, general logging facility.
 
 namespace eval ftp {
-
-namespace export DisplayMsg Open Close Cd Pwd Type List NList FileSize ModTime\
-		 Delete Rename Put Append Get Reget Newer Quote MkDir RmDir 
+    namespace export DisplayMsg Open Close Cd Pwd Type List NList \
+	    FileSize ModTime Delete Rename Put Append Get Reget \
+	    Newer Quote MkDir RmDir
 	
-set serial 0
-set VERBOSE 0
-set DEBUG 0
+    set serial 0
+    set VERBOSE 0
+    set DEBUG 0
 }
 
 #############################################################################
@@ -68,32 +83,27 @@ set DEBUG 0
 proc ftp::DisplayMsg {s msg {state ""}} {
 
     upvar ::ftp::ftp$s ftp
-    variable VERBOSE 
     
     if { ([info exists ftp(Output)]) && ($ftp(Output) != "") } {
         eval [concat $ftp(Output) {$s $msg $state}]
         return
     }
+
+    # FIX #476729. Instead of changing the documentation this
+    #              procedure is changed to enforce the documented
+    #              behaviour. IOW, this procedure will not throw
+    #              errors anymore. At the same time printing to stdout
+    #              is exchanged against calls into the 'log' module
+    #              tcllib, which is much easier to customize for the
+    #              needs of any application using the ftp module. The
+    #              variable VERBOSE is still relevant as it controls
+    #              whether this procedure is called or not.
         
     switch -exact -- $state {
-        data {
-            if { $VERBOSE } {
-                puts $msg
-            }
-        }
-        control	{
-            if { $VERBOSE } {
-                puts $msg
-            }
-        }
-        error {
-            error "ERROR: $msg"
-        }
-        default	{
-            if { $VERBOSE } {
-                puts $msg
-            }
-        }
+        data    {log::log debug "$state | $msg"}
+        control {log::log debug "$state | $msg"}
+        error   {log::log error "$state | E: $msg"}
+        default {log::log debug "$state | $msg"}
     }
     return
 }
@@ -270,8 +280,10 @@ proc ftp::StateHandler {s {sock ""}} {
             catch {unset ftp(State)}
             catch {close $ftp(CtrlSock); unset ftp(CtrlSock)}
             return
-        }
-	
+        } else {
+	    # Fix SF bug #466746: Incomplete line, do nothing.
+	    return	   
+	}
     } 
 	
     if { $DEBUG } {
@@ -895,7 +907,8 @@ proc ftp::StateHandler {s {sock ""}} {
                 2 {
                     set complete_with 1
 		    set nextState 1
-		    Command $ftp(Command) get $ftp(RemoteFilename)
+		    Command $ftp(Command) get $ftp(RemoteFilename):$ftp(From):$ftp(To)
+		    unset ftp(From) ftp(To)
                 }
                 default {
                     set errmsg "Error retrieving file \"$ftp(RemoteFilename)\"!"
@@ -1335,9 +1348,20 @@ proc ftp::ModTime {s {filename ""}} {
     }
 }
 
-proc ftp::ModTimePostProcess clock {
+proc ftp::ModTimePostProcess {clock} {
     foreach {year month day hour min sec} {1 1 1 1 1 1} break
-    scan $clock "%4s%2s%2s%2s%2s%2s" year month day hour min sec
+
+    # Bug #478478. Special code to detect ftp servers with a Y2K patch
+    # gone bad and delivering, hmmm, non-standard date information.
+
+    if {[string length $clock] == 15} {
+        scan $clock "%2s%3s%2s%2s%2s%2s%2s" cent year month day hour min sec
+        set year [expr {($cent * 100) + $year}]
+	log::log warning "data | W: server with non-standard time, bad Y2K patch."
+    } else {
+        scan $clock "%4s%2s%2s%2s%2s%2s" year month day hour min sec
+    }
+
     set clock [clock scan "$month/$day/$year $hour:$min:$sec" -gmt 1]
     return $clock
 }
@@ -1654,7 +1678,8 @@ proc ftp::Put {s args} {
         return 0
     }
     if {([llength $args] < 1) || ([llength $args] > 4)} {
-        DisplayMsg $s "wrong # args: should be \"ftp::Put handle (-data \"data\" | localFilename) remoteFilename\"" error
+        DisplayMsg $s \
+		"wrong # args: should be \"ftp::Put handle (-data \"data\" | -channel chan | localFilename) remoteFilename\"" error
 	return 0    
     }
 
@@ -1668,43 +1693,56 @@ proc ftp::Put {s args} {
         } elseif {($flags) && ([string equal $arg "-data"])} {
             set ftp(inline) 1
             set ftp(filebuffer) ""
+        } elseif {($flags) && ([string equal $arg "-channel"])} {
+            set ftp(inline) 2
 	} elseif {$source == ""} {
             set source $arg
 	} elseif {$dest == ""} {
             set dest $arg
 	} else {
-            DisplayMsg $s "wrong # args: should be \"ftp::Put handle (-data \"data\" | localFilename) remoteFilename\"" error
+            DisplayMsg $s "wrong # args: should be \"ftp::Put handle (-data \"data\" | -channel chan | localFilename) remoteFilename\"" error
 	    return 0
         }
     }
 
     if {($source == "")} {
-        DisplayMsg $s "Must specify a valid file to Put" error
+        DisplayMsg $s "Must specify a valid data source to Put" error
         return 0
     }        
 
     set ftp(RemoteFilename) $dest
 
-    if {$ftp(inline)} {
+    if {$ftp(inline) == 1} {
         set ftp(PutData) $source
         if { $dest == "" } {
             set dest ftp.tmp
         }
         set ftp(RemoteFilename) $dest
     } else {
-        set ftp(PutData) ""
-        if { ![file exists $source] } {
-            DisplayMsg $s "File \"$source\" not exist" error
-            return 0
-        }
-        if { $dest == "" } {
-            set dest [file tail $source]
-        }
-        set ftp(LocalFilename) $source
+	if {$ftp(inline) == 0} {
+	    # File transfer
+
+	    set ftp(PutData) ""
+	    if { ![file exists $source] } {
+		DisplayMsg $s "File \"$source\" not exist" error
+		return 0
+	    }
+	    if { $dest == "" } {
+		set dest [file tail $source]
+	    }
+	    set ftp(LocalFilename) $source
+	    set ftp(SourceCI) [open $ftp(LocalFilename) r]
+	} else {
+	    # Channel transfer. We fake the rest of the system into
+	    # believing that a file transfer is happening. This makes
+	    # the handling easier.
+
+	    set ftp(SourceCI) $source
+	    set ftp(inline) 0
+	}
         set ftp(RemoteFilename) $dest
 
 	# TODO: read from source file asynchronously
-        set ftp(SourceCI) [open $ftp(LocalFilename) r]
         if { [string equal $ftp(Type) "ascii"] } {
             fconfigure $ftp(SourceCI) -buffering line -blocking 1
         } else {
@@ -1756,7 +1794,7 @@ proc ftp::Append {s args} {
     }
 
     if {([llength $args] < 1) || ([llength $args] > 4)} {
-        DisplayMsg $s "wrong # args: should be \"ftp::Append handle (-data \"data\" | localFilename) remoteFilename\"" error
+        DisplayMsg $s "wrong # args: should be \"ftp::Append handle (-data \"data\" | -channel chan | localFilename) remoteFilename\"" error
         return 0
     }
 
@@ -1770,44 +1808,57 @@ proc ftp::Append {s args} {
         } elseif {($flags) && ([string equal $arg "-data"])} {
             set ftp(inline) 1
             set ftp(filebuffer) ""
+        } elseif {($flags) && ([string equal $arg "-channel"])} {
+            set ftp(inline) 2
         } elseif {$source == ""} {
             set source $arg
         } elseif {$dest == ""} {
             set dest $arg
         } else {
-            DisplayMsg $s "wrong # args: should be \"ftp::Append handle (-data \"data\" | localFilename) remoteFilename\"" error
+            DisplayMsg $s "wrong # args: should be \"ftp::Append handle (-data \"data\" | -channel chan | localFilename) remoteFilename\"" error
             return 0
         }
     }
 
     if {($source == "")} {
-        DisplayMsg $s "Must specify a valid file to Append" error
+        DisplayMsg $s "Must specify a valid data source to Append" error
         return 0
     }   
 
     set ftp(RemoteFilename) $dest
 
-    if {$ftp(inline)} {
+    if {$ftp(inline) == 1} {
         set ftp(PutData) $source
         if { $dest == "" } {
             set dest ftp.tmp
         }
         set ftp(RemoteFilename) $dest
     } else {
-        set ftp(PutData) ""
-        if { ![file exists $source] } {
-            DisplayMsg $s "File \"$source\" not exist" error
-            return 0
-        }
-			
-        if { $dest == "" } {
-            set dest [file tail $source]
-        }
+	if {$ftp(inline) == 0} {
+	    # File transfer
 
-        set ftp(LocalFilename) $source
+	    set ftp(PutData) ""
+	    if { ![file exists $source] } {
+		DisplayMsg $s "File \"$source\" not exist" error
+		return 0
+	    }
+			
+	    if { $dest == "" } {
+		set dest [file tail $source]
+	    }
+
+	    set ftp(LocalFilename) $source
+	    set ftp(SourceCI) [open $ftp(LocalFilename) r]
+	} else {
+	    # Channel transfer. We fake the rest of the system into
+	    # believing that a file transfer is happening. This makes
+	    # the handling easier.
+
+	    set ftp(SourceCI) $source
+	    set ftp(inline) 0
+	}
         set ftp(RemoteFilename) $dest
 
-        set ftp(SourceCI) [open $ftp(LocalFilename) r]
         if { [string equal $ftp(Type) "ascii"] } {
             fconfigure $ftp(SourceCI) -buffering line -blocking 1
         } else {
@@ -1858,7 +1909,7 @@ proc ftp::Get {s args} {
     }
 
     if {([llength $args] < 1) || ([llength $args] > 4)} {
-        DisplayMsg $s "wrong # args: should be \"ftp::Get handle remoteFile ?(-variable varName | localFilename)?\"" error
+        DisplayMsg $s "wrong # args: should be \"ftp::Get handle remoteFile ?(-variable varName | -channel chan | localFilename)?\"" error
 	return 0    
     }
 
@@ -1873,60 +1924,84 @@ proc ftp::Get {s args} {
         } elseif {($flags) && ([string equal $arg "-variable"])} {
             set ftp(inline) 1
             set ftp(filebuffer) ""
-	} elseif {($ftp(inline)) && ([string equal $varname "**NONE**"])} {
+        } elseif {($flags) && ([string equal $arg "-channel"])} {
+            set ftp(inline) 2
+	} elseif {($ftp(inline) == 1) && ([string equal $varname "**NONE**"])} {
             set varname $arg
 	    set ftp(get:varname) $varname
+	} elseif {($ftp(inline) == 2) && ([string equal $varname "**NONE**"])} {
+	    set ftp(get:channel) $arg
 	} elseif {$source == ""} {
             set source $arg
 	} elseif {$dest == ""} {
             set dest $arg
 	} else {
             DisplayMsg $s "wrong # args: should be \"ftp::Get handle remoteFile
-?(-variable varName | localFilename)?\"" error
+?(-variable varName | -channel chan | localFilename)?\"" error
 	    return 0
         }
     }
 
-    if {($ftp(inline)) && ($dest != "")} {
+    if {($ftp(inline) != 0) && ($dest != "")} {
         DisplayMsg $s "Cannot return data in a variable, and place it in destination file." error
         return 0
     }
 
     if {$source == ""} {
-        DisplayMsg $s "Must specify a valid file to Get" error
+        DisplayMsg $s "Must specify a valid data source to Get" error
         return 0
     }
 
-    if { $dest == "" } {
-        set dest $source
-    } else {
-        if {[file isdirectory $dest]} {
-            set dest [file join $dest [file tail $source]]
-        }
+    if {$ftp(inline) != 2} {
+	if { $dest == "" } {
+	    set dest $source
+	} else {
+	    if {[file isdirectory $dest]} {
+		set dest [file join $dest [file tail $source]]
+	    }
+	}
+	if {![file exists [file dirname $dest]]} {
+	    return -code error "ftp::Get, directory \"[file dirname $dest]\" for destination \"$dest\" does not exist"
+	}
+	set ftp(LocalFilename) $dest
     }
 
     set ftp(RemoteFilename) $source
-    set ftp(LocalFilename) $dest
 
+    if {$ftp(inline) == 2} {
+	set ftp(inline) 0
+    }
     set ftp(State) get_$ftp(Mode)
     StateHandler $s
 
     # wait for synchronization
     set rc [WaitOrTimeout $s]
+
+    # It is important to unset 'get:channel' in all cases or it will
+    # interfere with any following ftp command (as its existence
+    # suppresses the closing of the destination channel identifier
+    # (DestCI). We cannot do it earlier than just before the 'return'
+    # or code depending on it for the current command may not execute
+    # correctly.
+
     if { $rc } {
 	if {![string length $ftp(Command)]} {
 	    ElapsedTime $s [clock seconds]
 	    if {$ftp(inline)} {
+		catch {unset ftp(get:channel)}
 		upvar $varname returnData
 		set returnData $ftp(GetData)
 	    }
 	}
+	catch {unset ftp(get:channel)}
         return 1
     } else {
         if {$ftp(inline)} {
+	    catch {unset ftp(get:channel)}
             return ""
 	}
         CloseDataConn $s
+	catch {unset ftp(get:channel)}
         return 0
     }
 }
@@ -1948,9 +2023,9 @@ proc ftp::Get {s args} {
 # 0 -			file not retrieved
 # 1 - 			OK
 
-proc ftp::Reget {s source {dest ""}} {
+proc ftp::Reget {s source {dest ""} {from_bytes 0} {till_bytes -1}} {
     upvar ::ftp::ftp$s ftp
-
+    
     if { ![info exists ftp(State)] } {
         DisplayMsg $s "Not connected!" error
         return 0
@@ -1959,14 +2034,44 @@ proc ftp::Reget {s source {dest ""}} {
     if { $dest == "" } {
         set dest $source
     }
+    if {![file exists [file dirname $dest]]} {
+	return -code error "ftp::Reget, directory \"[file dirname $dest]\" for destination \"$dest\" does not exist"
+    }
 
     set ftp(RemoteFilename) $source
     set ftp(LocalFilename) $dest
+    set ftp(From) $from_bytes
+
+
+    # Assumes that the local file has a starting offset of $from_bytes
+    # The following calculation ensures that the download starts from the
+    # correct offset
 
     if { [file exists $ftp(LocalFilename)] } {
-        set ftp(FileSize) [file size $ftp(LocalFilename)]
+	set ftp(FileSize) [ expr [file size $ftp(LocalFilename)] + $from_bytes ]
+	 	
+	if { $till_bytes != -1 } {
+	    set ftp(To)   $till_bytes	
+	    set ftp(Bytes_to_go) [ expr $till_bytes - $ftp(FileSize) ]
+	
+	    if { $ftp(Bytes_to_go) <= 0 } return 0
+
+	} else {
+	    # till_bytes not set
+	    set ftp(To)   end
+	}
+
     } else {
-        set ftp(FileSize) 0
+	# local file does not exist
+        set ftp(FileSize) $from_bytes
+		  
+	if { $till_bytes != -1 } {
+	    set ftp(Bytes_to_go) [ expr $till_bytes - $from_bytes ]
+	    set ftp(To) $till_bytes
+	} else {
+	    #till_bytes not set
+	    set ftp(To)   end
+	}
     }
 	
     set ftp(State) reget_$ftp(Mode)
@@ -2019,6 +2124,9 @@ proc ftp::Newer {s source {dest ""}} {
     if { $dest == "" } {
         set dest $source
     }
+    if {![file exists [file dirname $dest]]} {
+	return -code error "ftp::Newer, directory \"[file dirname $dest]\" for destination \"$dest\" does not exist"
+    }
 
     set ftp(RemoteFilename) $source
     set ftp(LocalFilename) $dest
@@ -2228,7 +2336,8 @@ proc ftp::Open {server user passwd args} {
     if { $DEBUG } {
         DisplayMsg $s "Starting new connection with: "
     }
-	
+
+    set ftp(inline) 	0
     set ftp(User)       $user
     set ftp(Passwd) 	$passwd
     set ftp(RemoteHost) $server
@@ -2321,6 +2430,20 @@ proc ftp::CopyNext {s bytes {error {}}} {
 
     incr ftp(Total) $bytes
 
+    # update bytes_to_go and blocksize
+
+    if { [info exists ftp(Bytes_to_go)] } {
+	set ftp(Bytes_to_go) [expr $ftp(Bytes_to_go) - $bytes]
+	 
+	if { $ftp(Blocksize) <= $ftp(Bytes_to_go) } {
+	    set blocksize $ftp(Blocksize)
+	} else {
+	    set blocksize $ftp(Bytes_to_go)
+	}
+    } else {
+	set blocksize $ftp(Blocksize)
+    } 
+    
     # callback for progress bar procedure
     
     if { ([info exists ftp(Progress)]) && \
@@ -2339,13 +2462,23 @@ proc ftp::CopyNext {s bytes {error {}}} {
     }
 
     if { $error != "" } {
-        catch {close $ftp(DestCI)}
+	# Protect the destination channel from destruction if it came
+	# from the caller. Closing it is not our responsibility in that case.
+
+	if {![info exists ftp(get:channel)]} {
+	    catch {close $ftp(DestCI)}
+	}
         catch {close $ftp(SourceCI)}
         unset ftp(state.data)
         DisplayMsg $s $error error
 
-    } elseif { [eof $ftp(SourceCI)] } {
-        close $ftp(DestCI)
+    } elseif { ([eof $ftp(SourceCI)] || ($blocksize <= 0)) } {
+	# Protect the destination channel from destruction if it came
+	# from the caller. Closing it is not our responsibility in that case.
+
+	if {![info exists ftp(get:channel)]} {
+	    close $ftp(DestCI)
+	}
         close $ftp(SourceCI)
         unset ftp(state.data)
         if { $VERBOSE } {
@@ -2353,8 +2486,9 @@ proc ftp::CopyNext {s bytes {error {}}} {
         }
 
     } else {
-        fcopy $ftp(SourceCI) $ftp(DestCI) -command [list [namespace current]::CopyNext $s] -size $ftp(Blocksize)
-
+	fcopy $ftp(SourceCI) $ftp(DestCI) \
+		-command [list [namespace current]::CopyNext $s] \
+		-size $blocksize
     }
     return
 }
@@ -2379,7 +2513,16 @@ proc ftp::HandleData {s sock} {
     # create local file for ftp::Get 
 
     if { [regexp -- "^get" $ftp(State)]  && (!$ftp(inline))} {
-        set rc [catch {set ftp(DestCI) [open $ftp(LocalFilename) w]} msg]
+
+	# A channel was specified by the caller. Use that instead of a
+	# file.
+
+	if {[info exists ftp(get:channel)]} {
+	    set ftp(DestCI) $ftp(get:channel)
+	    set rc 0
+	} else {
+	    set rc [catch {set ftp(DestCI) [open $ftp(LocalFilename) w]} msg]
+	}
         if { $rc != 0 } {
             DisplayMsg $s "$msg" error
             return 0
@@ -2408,11 +2551,28 @@ proc ftp::HandleData {s sock} {
         }
     }	
 
-    # perform fcopy
 
     set ftp(Total) 0
     set ftp(Start_Time) [clock seconds]
-    fcopy $ftp(SourceCI) $ftp(DestCI) -command [list [namespace current]::CopyNext $s] -size $ftp(Blocksize)
+	 
+    # calculate blocksize
+	 
+    if { [ info exists ftp(Bytes_to_go) ] } {
+			
+	if { $ftp(Blocksize) <= $ftp(Bytes_to_go) } {
+	    set Blocksize $ftp(Blocksize)
+	} else {
+	    set Blocksize $ftp(Bytes_to_go)
+	}
+	
+    } else {
+	set Blocksize $ftp(Blocksize)
+    }
+	
+    # perform fcopy
+    fcopy $ftp(SourceCI) $ftp(DestCI) \
+	    -command [list [namespace current]::CopyNext $s ] \
+	    -size $Blocksize
     return 1
 }
 
@@ -2534,6 +2694,14 @@ proc ftp::HandleOutput {s sock} {
 proc ftp::CloseDataConn {s } {
     upvar ::ftp::ftp$s ftp
 
+    # Protect the destination channel from destruction if it came
+    # from the caller. Closing it is not our responsibility.
+
+    if {[info exists ftp(get:channel)]} {
+	catch {unset ftp(get:channel)}
+	catch {unset ftp(DestCI)}
+    }
+
     catch {after cancel $ftp(Wait)}
     catch {fileevent $ftp(DataSock) readable {}}
     catch {close $ftp(DataSock); unset ftp(DataSock)}
@@ -2765,15 +2933,16 @@ proc ftp::Command {cb msg args} {
     }
 }
 
+# ==================================================================
 # ?????? Hmm, how to do multithreaded for tkcon?
 # added TkCon support
-# TkCon is (c) 1995-1999 Jeffrey Hobbs, http://www.purl.org/net/hobbs/tcl/script/tkcon/
+# TkCon is (c) 1995-2001 Jeffrey Hobbs, http://tkcon.sourceforge.net/
 # started with: tkcon -load ftp
 if { [string equal [uplevel "#0" {info commands tkcon}] "tkcon"] } {
 
     # new ftp::List proc makes the output more readable
     proc ::ftp::__ftp_ls {args} {
-        foreach i [::ftp::List_org $args] {
+        foreach i [eval ::ftp::List_org $args] {
             puts $i
         }
     }
@@ -2788,3 +2957,7 @@ if { [string equal [uplevel "#0" {info commands tkcon}] "tkcon"] } {
     set ::ftp::DEBUG 0
 }
 
+# ==================================================================
+# At last, everything is fine, we can provide the package.
+
+package provide ftp [lindex {Revision: 2.3} 1]
