@@ -9,11 +9,12 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 # 
-# RCS: @(#) $Id: ftpd.tcl,v 1.8 2001/06/22 15:29:18 andreas_kupries Exp $
+# RCS: @(#) $Id: ftpd.tcl,v 1.12 2001/09/24 19:05:14 jenglish Exp $
 #
 
 # Define the ftpd package version 1.1.1
 
+package require Tcl 8.2
 namespace eval ::ftpd {
 
     # The listening port.
@@ -102,7 +103,8 @@ namespace eval ::ftpd {
 
     # Exported procedures
 
-    namespace export *
+    namespace export config hasCallback logStderr 
+    namespace export fileAuth anonAuth unixAuth server accept read
 }
 
 # ::ftpd::config --
@@ -652,7 +654,7 @@ proc ::ftpd::Fs {command path args} {
 # ftp commands with the same name as the procs.
 
 namespace eval ::ftpd::command {
-    namespace export *
+    # All commands in this namespace are private, no export.
 }
 
 # ::ftpd::command::ABOR --
@@ -718,7 +720,7 @@ proc ::ftpd::command::APPE {sock list} {
 	puts $sock "150 Copy Started ($data(mode))"
 	fcopy $data(sock2) $f -command [list ::ftpd::GetDone $sock $data(sock2) $f ""]
     } else {
-	puts $sock "500 Copy Failed: $path\n$f"
+	puts $sock "500 Copy Failed: $path $f"
 	::ftpd::FinishData $sock
     }
     return
@@ -1198,7 +1200,7 @@ proc ::ftpd::command::RETR {sock list} {
 	puts $sock "150 Copy Started ($data(mode))"
 	fcopy $f $data(sock2) -command [list ::ftpd::GetDone $sock $data(sock2) $f ""]
     } else {
-	puts $sock "500 Copy Failed: $path\n$f"
+	puts $sock "500 Copy Failed: $path $f"
 	::ftpd::FinishData $sock
     }
     return
@@ -1390,7 +1392,7 @@ proc ::ftpd::command::STOR {sock list} {
 	puts $sock "150 Copy Started ($data(mode))"
 	fcopy $data(sock2) $f -command [list ::ftpd::GetDone $sock $data(sock2) $f ""]
     } else {
-	puts $sock "500 Copy Failed: $path\n$f"
+	puts $sock "500 Copy Failed: $path $f"
 	::ftpd::FinishData $sock
     }
     return
@@ -1441,7 +1443,7 @@ proc ::ftpd::command::STOU {sock list} {
 	puts $sock "150 Copy Started ($data(mode))"
 	fcopy $data(sock2) $f -command [list ::ftpd::GetDone $sock $data(sock2) $f $file]
     } else {
-	puts $sock "500 Copy Failed: $path\n$f"
+	puts $sock "500 Copy Failed: $path $f"
 	::ftpd::FinishData $sock
     }
     return
@@ -1621,7 +1623,7 @@ namespace eval ::ftpd::fsFile {
 	set docRoot /
     }
 
-    namespace export *
+    namespace export docRoot fs
 }
 
 # ::ftpd::fsFile::docRoot --
