@@ -9,8 +9,10 @@
 # unpublished package of 1999.
 #
 
+# new string features and inline scan are used, requiring 8.3.
+package require Tcl 8.3
 
-package provide mime 1.2
+package provide mime 1.3.1
 
 if {[catch {package require Trf  2.0}]} {
 
@@ -316,7 +318,7 @@ proc mime::initializeaux {token args} {
 
             -encoding {
                 switch -- [set state(encoding) [string tolower $value]] {
-                    7bit - 8bit - quoted-printable - base64 {
+                    7bit - 8bit - binary - quoted-printable - base64 {
                     }
 
                     default {
@@ -1307,7 +1309,8 @@ proc mime::getbody {token args} {
                             set chunk [$state(encoding) -mode decode \
                                                         -- $chunk]
                         }
-			"" {
+			7bit - 8bit - binary - "" {
+			    # Bugfix for [#477088]
 			    # Go ahead, leave chunk alone
 			}
 			default {
@@ -1373,6 +1376,8 @@ proc mime::getbody {token args} {
                 }
 
                 default {
+		    # Not a bugfix for [#477088], but clarification
+		    # This handles no-encoding, 7bit, 8bit, and binary.
                     set fragment $state(string)
                 }
             }
@@ -1558,7 +1563,8 @@ proc mime::copymessageaux {token channel} {
                 quoted-printable {
                     set converter $encoding
                 }
-		"" {
+		7bit - 8bit - binary - "" {
+		    # Bugfix for [#477088]
 		    # Go ahead
 		}
 		default {
@@ -1796,7 +1802,8 @@ proc mime::buildmessageaux {token} {
                 quoted-printable {
                     set converter $encoding
                 }
-		"" {
+		7bit - 8bit - binary - "" {
+		    # Bugfix for [#477088]
 		    # Go ahead
 		}
 		default {
@@ -2169,71 +2176,59 @@ proc mime::scopy {token channel offset len blocksize} {
 # Results:
 #	The properly quoted string is returned.
 
-proc mime::qp_encode {string {encoded_word 0}} {
+proc mime::qp_encode {string {encoded_word 0} {no_softbreak 0}} {
+    # 8.1+ improved string manipulation routines used.
+    # Replace outlying characters, characters that would normally
+    # be munged by EBCDIC gateways, and special Tcl characters "[\]{}
+    # with =xx sequence
 
-    # Replace outlying characters, characters that would normally be munged
-    # by EBCDIC gateways, and special Tcl characters with =xx sequence
-
-    # Backslash special Tcl characters
-    # \x22 == "
-    # \x5B == [
-    # \x5C == "\"
-    # \x5D == ]
-    # \x7B == "{"
-    # \x7D == "}"
-
-    regsub -all -- {[\x22=\x5B-\x5D\x7B\x7D]} \
-	$string {[format =%02X [scan "\\&" %c]]} string
-
-    # Replace outlying characters that would be munged by EBCDIC gateways.
-
-    regsub -all -- {[\x00-\x08\x0B-\x1E\x21\x23\x24\x40\x5E\x60\x7C\x7E-\xFF]} \
-	$string {[format =%02X [scan {&} %c]]} string
-
-    # soft/hard newlines
-
-    regsub -all -- {([ \t])\n} $string \
-	    {[scan "\1" %c c; format "=%02X\n" $c]} string
-
-    # Funky cases for SMTP compatibility
-
-    regsub -all -- "\n\.\n" $string "\n=2E\n" string
-    regsub -all -- "\nFrom " $string "\n=46rom " string
+    regsub -all -- \
+	    {[\x00-\x08\x0B-\x1E\x21-\x24\x40\x5B-\x5E\x60\x7B-\xFF]} \
+	    $string {[format =%02X [scan "\\&" %c]]} string
 
     # Replace the format commands with their result
 
     set string [subst -novariable $string]
 
-    # Special processing for encoded words (RFC 2047)
-    
+    # soft/hard newlines and other
+    # Funky cases for SMTP compatibility
+    set mapChars [list " \n" "=20\n" "\t\n" "=09\n" \
+	    "\n\.\n" "\n=2E\n" "\nFrom " "\n=46rom "]
     if {$encoded_word} {
-        regsub -all -- " " $string "_" string
+	# Special processing for encoded words (RFC 2047)
+	lappend mapChars " " "_"
     }
+    set string [string map $mapChars $string]
 
     # Break long lines - ugh
 
-    set result ""
-    foreach line [split $string \n] {
-	while {[string length $line] > 72} {
-	    set chunk [string range $line 0 72]
-	    if {[regexp -- (=|=.)$ $chunk dummy end]} {
+    # Implementation of FR #503336
+    if {$no_softbreak} {
+	set result $string
+    } else {
+	set result ""
+	foreach line [split $string \n] {
+	    while {[string length $line] > 72} {
+		set chunk [string range $line 0 72]
+		if {[regexp -- (=|=.)$ $chunk dummy end]} {
+		    
+		    # Don't break in the middle of a code
 
-		# Don't break in the middle of a code
-
-		set len [expr {72 - [string length $end]}]
-		set chunk [string range $line 0 $len]
-		incr len
-		set line [string range $line $len end]
-	    } else {
-		set line [string range $line 73 end]
+		    set len [expr {72 - [string length $end]}]
+		    set chunk [string range $line 0 $len]
+		    incr len
+		    set line [string range $line $len end]
+		} else {
+		    set line [string range $line 73 end]
+		}
+		append result $chunk=\n
 	    }
-	    append result $chunk=\n
+	    append result $line\n
 	}
-	append result $line\n
     }
     
-    # Trim off last \n, since the above code has the side-effect of adding an
-    # extra \n to the encoded string and return the result.
+    # Trim off last \n, since the above code has the side-effect
+    # of adding an extra \n to the encoded string and return the result.
 
     set result [string range $result 0 end-1]
 
@@ -2242,11 +2237,10 @@ proc mime::qp_encode {string {encoded_word 0}} {
     set lastChar [string index $result end]
     if {$lastChar==" "} {
 	set result [string replace $result end end "=20"]
-    }
-    if {$lastChar=="\t"} {
+    } elseif {$lastChar=="\t"} {
 	set result [string replace $result end end "=09"]
     }
-	
+
     return $result
 }
 
@@ -2263,32 +2257,33 @@ proc mime::qp_encode {string {encoded_word 0}} {
 #	The decoded string is returned.
 
 proc mime::qp_decode {string {encoded_word 0}} {
-
+    # 8.1+ improved string manipulation routines used.
     # Special processing for encoded words (RFC 2047)
 
     if {$encoded_word} {
-        regsub -all -- "_" $string " " string
+	# _ == \x20, even if SPACE occupies a different code position
+	set string [string map [list _ \u0020] $string]
     }
 
-    # Protect Tcl special chars
-
-    regsub -all -- {[][\\\$]} $string {\\&} string
-
     # smash the white-space at the ends of lines since that must've been
     # generated by an MUA.
 
     regsub -all -- {[ \t]+\n} $string "\n" string
     set string [string trimright $string " \t"]
 
-    # smash soft newlines
+    # Protect the backslash for later subst and
+    # smash soft newlines, has to occur after white-space smash
+    # and any encoded word modification.
 
-    regsub -all -- {=\n} $string {} string
+    set string [string map [list "\\" "\\\\" "=\n" ""] $string]
 
     # Decode specials
 
-    regsub -all -- =(..) $string {[format %c "0x\1"]} string
+    regsub -all -nocase {=([a-f0-9][a-f0-9])} $string {\\u00\1} string
 
-    return [subst $string]
+    # process \u unicode mapped chars
+
+    return [subst -novar -nocommand $string]
 }
 
 # mime::parseaddress --
