@@ -7,7 +7,7 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
-# RCS: @(#) $Id: stats.tcl,v 1.16 2001/07/10 20:39:47 andreas_kupries Exp $
+# RCS: @(#) $Id: stats.tcl,v 1.17 2001/09/05 23:53:52 andreas_kupries Exp $
 
 package provide stats 1.0
 
@@ -38,7 +38,9 @@ namespace eval stats:: {
     }
     variable mergeInterval
 
-    namespace export *
+    namespace export countInit countReset count countExists countGet countNames
+    namespace export histHtmlDisplay histHtmlDisplayRow histHtmlDisplayBarChart
+    namespace export countStart countStop
 }
 
 # stats::countInit --
